@@ -59,6 +59,9 @@
 double rsvg_internal_dpi_x = RSVG_DEFAULT_DPI_X;
 double rsvg_internal_dpi_y = RSVG_DEFAULT_DPI_Y;
 
+static xmlSAXHandler rsvgSAXHandlerStruct;
+static gboolean rsvgSAXHandlerStructInited = FALSE;
+
 typedef struct _RsvgSaxHandlerDefs {
 	RsvgSaxHandler super;
 	RsvgHandle *ctx;
@@ -353,7 +356,7 @@ rsvg_title_handler_free (RsvgSaxHandler *self)
 static void
 rsvg_title_handler_characters (RsvgSaxHandler *self, const char *ch, int len)
 {
-	RsvgSaxHandlerDesc *z = (RsvgSaxHandlerDesc *)self;
+	RsvgSaxHandlerTitle *z = (RsvgSaxHandlerTitle *)self;
 	RsvgHandle *ctx = z->ctx;
 
 	/* This isn't quite the correct behavior - in theory, any graphics
@@ -503,13 +506,156 @@ rsvg_start_metadata (RsvgHandle *ctx, RsvgPropertyBag *atts)
 
 /* end metadata */
 
+/* start xinclude */
+
+typedef struct _RsvgSaxHandlerXinclude {
+	RsvgSaxHandler super;
+
+	RsvgSaxHandler *prev_handler;
+	RsvgHandle *ctx;
+	gboolean success;
+	gboolean in_fallback;
+} RsvgSaxHandlerXinclude;
+
+static void
+rsvg_start_xinclude (RsvgHandle *ctx, RsvgPropertyBag *atts);
+static void
+rsvg_characters_impl (RsvgHandle *ctx, const xmlChar * ch, int len);
+
+static void
+rsvg_xinclude_handler_free (RsvgSaxHandler *self)
+{
+	g_free (self);
+}
+
+static void
+rsvg_xinclude_handler_characters (RsvgSaxHandler *self, const char *ch, int len)
+{
+	RsvgSaxHandlerXinclude *z = (RsvgSaxHandlerXinclude *)self;
+
+	if (z->in_fallback) {
+		rsvg_characters_impl (z->ctx, (const xmlChar *)ch, len);
+	}
+}
+
+static void
+rsvg_xinclude_handler_start (RsvgSaxHandler *self, const char *name,
+							 RsvgPropertyBag *atts)
+{
+	RsvgSaxHandlerXinclude *z = (RsvgSaxHandlerXinclude *)self;
+
+	if (!z->success) {
+		if (z->in_fallback) {
+			if(!strcmp(name, "xi:include"))
+				rsvg_start_xinclude (z->ctx, atts);
+			else
+				rsvg_standard_element_start (z->ctx, (const char *)name, atts);
+		} else if (!strcmp (name, "xi:fallback")) {
+			z->in_fallback = TRUE;
+		}
+	}
+}
+
+static void
+rsvg_xinclude_handler_end (RsvgSaxHandler *self, const char *name)
+{
+	RsvgSaxHandlerXinclude *z = (RsvgSaxHandlerXinclude *)self;
+	RsvgHandle *ctx = z->ctx;
+	
+	if (!strcmp(name, "include") || !strcmp(name, "xi:include"))
+		{
+			if (ctx->priv->handler != NULL)
+				{
+					RsvgSaxHandler * previous_handler;
+
+					previous_handler = z->prev_handler;
+					ctx->priv->handler->free (ctx->priv->handler);
+					ctx->priv->handler = previous_handler;
+				}
+		}
+	else if (z->in_fallback) {
+		if (!strcmp(name, "xi:fallback"))
+			z->in_fallback = FALSE;
+	}
+}
+
+/* http://www.w3.org/TR/xinclude/ */
+static void
+rsvg_start_xinclude (RsvgHandle *ctx, RsvgPropertyBag *atts)
+{
+	RsvgSaxHandlerXinclude *handler;
+	GByteArray * data;
+	const char * href;
+	gboolean success = FALSE;
+
+	href = rsvg_property_bag_lookup (atts, "href");
+	if (href) {
+		data = _rsvg_acquire_xlink_href_resource (href, rsvg_handle_get_base_uri (ctx), NULL);
+		if (data) {
+			const char * parse;
+			
+			parse = rsvg_property_bag_lookup (atts, "parse");
+			if (parse && !strcmp (parse, "text")) {
+				const char * encoding;
+				char * text_data;
+				gsize text_data_len;
+				gboolean text_data_needs_free = FALSE;
+				
+				encoding = rsvg_property_bag_lookup (atts, "encoding");
+				if (encoding) {
+					text_data = g_convert ((const char *)data->data, data->len, "utf-8", encoding, NULL, &text_data_len, NULL);
+					text_data_needs_free = TRUE;
+				} else {
+					text_data = (char *)data->data;
+					text_data_len = data->len;
+				}
+				
+				rsvg_characters_impl (ctx, (const xmlChar *)text_data, text_data_len);
+				
+				if (text_data_needs_free)
+					g_free (text_data);
+			} else {
+				/* xml */
+				xmlDocPtr xml_doc;
+				xmlParserCtxtPtr xml_parser;
+				int result;
+				
+				xml_parser = xmlCreatePushParserCtxt (&rsvgSAXHandlerStruct, ctx, NULL, 0, NULL);
+				result = xmlParseChunk (xml_parser, (char*)data->data, data->len, 0);
+				result = xmlParseChunk (xml_parser, "", 0, TRUE);
+				
+				xml_doc = xml_parser->myDoc;
+				xmlFreeParserCtxt (xml_parser);
+				xmlFreeDoc(xml_doc);
+			}
+			
+			g_byte_array_free (data, TRUE);
+			success = TRUE;
+		}
+	}
+
+	/* needed to handle xi:fallback */
+	handler = g_new0 (RsvgSaxHandlerXinclude, 1);
+	
+	handler->super.free = rsvg_xinclude_handler_free;
+	handler->super.characters = rsvg_xinclude_handler_characters;
+	handler->super.start_element = rsvg_xinclude_handler_start;
+	handler->super.end_element   = rsvg_xinclude_handler_end;
+	handler->prev_handler = ctx->priv->handler;
+	handler->ctx = ctx;
+	handler->success = success;
+
+	ctx->priv->handler = &handler->super;
+}
+
+/* end xinclude */
+
 static void
 rsvg_start_element (void *data, const xmlChar *name,
 					const xmlChar ** atts)
 {
-	RsvgHandle *ctx = (RsvgHandle *)data;
-
 	RsvgPropertyBag * bag;
+	RsvgHandle *ctx = (RsvgHandle *)data;
 
 	bag = rsvg_property_bag_new((const char **)atts);
 
@@ -534,7 +680,10 @@ rsvg_start_element (void *data, const xmlChar *name,
 				rsvg_start_desc (ctx, bag);
 			else if (!strcmp ((const char *)name, "metadata"))
 				rsvg_start_metadata (ctx, bag);
-			rsvg_standard_element_start (ctx, (const char *)name, bag);
+			else if (!strcmp ((const char *)name, "include")) /* xi:include */
+				rsvg_start_xinclude (ctx, bag);
+			else
+				rsvg_standard_element_start (ctx, (const char *)name, bag);
     }
 
 	rsvg_property_bag_free(bag);
@@ -573,16 +722,9 @@ static void _rsvg_node_chars_free(RsvgNode * node)
 }
 
 static void
-rsvg_characters (void *data, const xmlChar *ch, int len)
+rsvg_characters_impl (RsvgHandle *ctx, const xmlChar * ch, int len)
 {
-	RsvgHandle *ctx = (RsvgHandle *)data;
 	RsvgNodeChars * self;
-	
-	if (ctx->priv->handler && ctx->priv->handler->characters != NULL)
-		{
-			ctx->priv->handler->characters (ctx->priv->handler, (const char *)ch, len);
-			return;
-		}
 
 	if (!ch || !len)
 		return;
@@ -611,32 +753,43 @@ rsvg_characters (void *data, const xmlChar *ch, int len)
 		rsvg_node_group_pack(ctx->priv->currentnode, (RsvgNode *)self);
 }
 
+static void
+rsvg_characters (void *data, const xmlChar *ch, int len)
+{
+	RsvgHandle *ctx = (RsvgHandle *)data;
+	
+	if (ctx->priv->handler && ctx->priv->handler->characters != NULL)
+		{
+			ctx->priv->handler->characters (ctx->priv->handler, (const char *)ch, len);
+			return;
+		}
+	
+	rsvg_characters_impl (ctx, ch, len);
+}
+
 #if LIBXML_VERSION >= 20621
 #define RSVG_ENABLE_ENTITIES
 #elif defined(__GNUC__)
 #warning "libxml version less than 2.6.22. XML entities won't work"
 #endif
 
+#ifdef RSVG_ENABLE_ENTITIES
+
 static xmlEntityPtr
 rsvg_get_entity (void *data, const xmlChar *name)
 {
-#ifdef RSVG_ENABLE_ENTITIES
 	RsvgHandle *ctx = (RsvgHandle *)data;
 	xmlEntityPtr entity;
 
 	entity = g_hash_table_lookup (ctx->priv->entities, name);
 
 	return entity;
-#else
-	return NULL;
-#endif
 }
 
 static void
 rsvg_entity_decl (void *data, const xmlChar *name, int type,
 				  const xmlChar *publicId, const xmlChar *systemId, xmlChar *content)
 {
-#ifdef RSVG_ENABLE_ENTITIES
 	RsvgHandle *ctx = (RsvgHandle *)data;
 	GHashTable *entities = ctx->priv->entities;
 	xmlEntityPtr entity;
@@ -644,20 +797,52 @@ rsvg_entity_decl (void *data, const xmlChar *name, int type,
 
 	entity = g_new0 (xmlEntity, 1);
 	entity->type = XML_ENTITY_DECL;
-	dupname = (xmlChar *) g_strdup ((char *)name);
-	entity->name = dupname;
-	entity->ExternalID = (xmlChar *) g_strdup ((char *)publicId);
-	entity->SystemID = (xmlChar *) g_strdup ((char *)systemId);
+	dupname = (xmlChar *) g_strdup ((const char *)name);
+	entity->name = dupname;	
+	
 	entity->etype = type;
 	if (content)
 		{
-			entity->content = (xmlChar *) xmlMemStrdup ((char *)content);
-			entity->length = strlen ((char *)content);
+			entity->content = (xmlChar *) xmlMemStrdup ((const char *)content);
+			entity->length = strlen ((const char *)content);
 		}
+	else if (systemId || publicId)
+		{
+			GByteArray * data = NULL;
+
+			if (systemId)
+				data = _rsvg_acquire_xlink_href_resource ((const char *)systemId, rsvg_handle_get_base_uri (ctx), NULL);
+			else if (publicId)
+				data = _rsvg_acquire_xlink_href_resource ((const char *)publicId, rsvg_handle_get_base_uri (ctx), NULL);
+
+			if (data) {
+				entity->SystemID = (xmlChar *) g_strdup ((const char *)systemId);
+				entity->ExternalID = (xmlChar *) g_strdup ((const char *)publicId);
+				entity->content = (xmlChar *) xmlMemStrdup ((const char *)data->data);
+				entity->length = data->len;
+
+				/* fool libxml2 into supporting SYSTEM and PUBLIC entities */
+				entity->etype = XML_INTERNAL_GENERAL_ENTITY;
+
+				g_byte_array_free (data, TRUE);
+			}
+		}
+
 	g_hash_table_insert (entities, dupname, entity);
-#endif
 }
 
+static void
+rsvg_unparsed_entity_decl(void *ctx,
+						  const xmlChar *name,
+						  const xmlChar *publicId,
+						  const xmlChar *systemId,
+						  const xmlChar *notationName)
+{
+	rsvg_entity_decl(ctx, name, XML_INTERNAL_GENERAL_ENTITY, publicId, systemId, NULL);
+}
+
+#endif
+
 static void
 rsvg_error_cb (void *data, const char *msg, ...)
 {
@@ -670,8 +855,46 @@ rsvg_error_cb (void *data, const char *msg, ...)
 #endif
 }
 
-static xmlSAXHandler rsvgSAXHandlerStruct;
-static gboolean rsvgSAXHandlerStructInited = FALSE;
+static void
+rsvg_processing_instruction (void *ctx,
+							 const xmlChar *target,
+							 const xmlChar *data)
+{
+	/* http://www.w3.org/TR/xml-stylesheet/ */
+	RsvgHandle *handle = (RsvgHandle *)ctx;
+
+	if(!strcmp((const char *)target, "xml-stylesheet")) {
+		RsvgPropertyBag * atts;
+		char ** xml_atts;
+
+		xml_atts = rsvg_css_parse_xml_attribute_string((const char *)data);
+		atts = rsvg_property_bag_new ((const char **)xml_atts);
+
+		if(atts) {
+			const char * value;
+
+			value = rsvg_property_bag_lookup (atts, "alternate");
+			if (!value || (strcmp(value, "no") != 0)) {
+				value = rsvg_property_bag_lookup (atts, "type");
+				if (value && strcmp(value, "text/css") == 0) {
+					value = rsvg_property_bag_lookup (atts, "href");
+					if (value) {
+						GByteArray *style;
+					
+						style = _rsvg_acquire_xlink_href_resource (value, rsvg_handle_get_base_uri (handle), NULL);
+						if (style) {
+							rsvg_parse_cssbuffer (handle, (char *)style->data, style->len);
+							g_byte_array_free (style, TRUE);
+						}
+					}
+				}
+			}
+
+			g_strfreev (xml_atts);
+			rsvg_property_bag_free (atts);
+		}
+	}
+}
 
 void rsvg_SAX_handler_struct_init (void)
 {
@@ -681,13 +904,17 @@ void rsvg_SAX_handler_struct_init (void)
 
 			memset(&rsvgSAXHandlerStruct, 0, sizeof(rsvgSAXHandlerStruct));
 
+#ifdef RSVG_ENABLE_ENTITIES
 			rsvgSAXHandlerStruct.getEntity = rsvg_get_entity;
 			rsvgSAXHandlerStruct.entityDecl = rsvg_entity_decl;
+			rsvgSAXHandlerStruct.unparsedEntityDecl = rsvg_unparsed_entity_decl;
+#endif
 			rsvgSAXHandlerStruct.characters = rsvg_characters;
 			rsvgSAXHandlerStruct.error = rsvg_error_cb;
 			rsvgSAXHandlerStruct.cdataBlock = rsvg_characters;
 			rsvgSAXHandlerStruct.startElement = rsvg_start_element;
 			rsvgSAXHandlerStruct.endElement = rsvg_end_element;
+			rsvgSAXHandlerStruct.processingInstruction = rsvg_processing_instruction;
 		}
 }
 
@@ -780,8 +1007,9 @@ rsvg_handle_write_impl (RsvgHandle    *handle,
 	handle->priv->error = &real_error;
 	if (handle->priv->ctxt == NULL)
 		{
-			handle->priv->ctxt = xmlCreatePushParserCtxt (&rsvgSAXHandlerStruct, handle, NULL, 0, NULL);
-			handle->priv->ctxt->replaceEntities = TRUE;
+			handle->priv->ctxt = xmlCreatePushParserCtxt (&rsvgSAXHandlerStruct, handle, NULL, 0, 
+														  rsvg_handle_get_base_uri (handle));
+			/* handle->priv->ctxt->replaceEntities = TRUE; */
 		}
 	
 	result = xmlParseChunk (handle->priv->ctxt, (char*)buf, count, 0);
